/****************************************************************************

  PROGRAM: TestMouseCapture.C

  Scenario:
    Two windows A and B exist.
    - Window B captures the mouse input.
    - Window A opens the CommonFileDialog (GetOpenFileName,GetSaveFileName)
    - during the open process of the CommonFileDialog the mouse capture
      is released on Windows (between the 2nd & the 3rd WM_NOTIFY being
      passed to the OFNHookProc).

      The dialog is automatically closed by the hook proc, by posting
      WM_COMMAND with IDCANCEL.

      On Wine the mouse capture is not release, making it impossible
      to operate the dialog with the mouse.

      The same scenario may happen with ChooseFont (and probably
      more dialog windows).

      log output on Windows XP:
		Testing GetOpenFileName
		Wnd with capture:000201fe
		WM_INITDIALOG:000201fe
		WM_NOTIFY:000201fe - fffffda7 - (0)
		WM_NOTIFY:000201fe - fffffda5 - (1)
		WM_NOTIFY:00000000 - fffffda6 - (2)
		Quitting dialog
		Wnd with capture:00000000
		Testing ChooseFont
		Wnd with capture:000201fe
		WM_INITDIALOG:000201fe
		WM_CTLCOLORDLG:00000000 - (0)
		Quitting dialog
		Wnd with capture:00000000

      log output on Wine 1.1.0 (for OpenSuse 10.3)
		Testing GetOpenFileName
		Wnd with capture:0001002a
		WM_INITDIALOG:0001002a
		WM_NOTIFY:0001002a - fffffda7 - (0)
		WM_NOTIFY:0001002a - fffffda5 - (1)
		WM_NOTIFY:0001002a - fffffda6 - (2)
		CommonFileDialog - mouse capture should have been released now!
		Quitting dialog
		Wnd with capture:0001002a
		Testing ChooseFont
		Wnd with capture:0001002a
		WM_INITDIALOG:0001002a
		WM_CTLCOLORDLG:0001002a - (0)
		ChooseFontDialog - mouse capture should have been released now!
		Quitting dialog
		Wnd with capture:0001002a


****************************************************************************/

// #include <wine/test.h>
#include <windows.h>
#include <commdlg.h>
#include <string.h>
#include <stdlib.h>

#define BUFFERSIZE 512

HANDLE hInst;
HWND MAINWND;  /* handle to main window */
HWND CAPTWND;
HWND LOGWND;
INT  COUNT=0;

VOID WINAPI AppendToLog(CHAR *text)
	{
  	INT textLength, requiredSize, sizeBuffer, numDiscard ;
  	CHAR *buffer;

  	if (LOGWND != NULL)
  		{
  		textLength = GetWindowTextLength(LOGWND);
    	requiredSize = textLength + strlen(text) + 1;
    	sizeBuffer = requiredSize;
    	if ( sizeBuffer > BUFFERSIZE )
    		{
      		sizeBuffer = BUFFERSIZE;
    		}
    	buffer = malloc(sizeBuffer);
    	if (buffer)
    		{
      		GetWindowText(LOGWND,buffer,sizeBuffer);
      		if (requiredSize >= BUFFERSIZE)
      			{
        		numDiscard = requiredSize - BUFFERSIZE + 1;
        		strcpy(buffer, buffer + numDiscard );
      			}
      		strcat(buffer,text);
      		SetWindowText(LOGWND,buffer);
      		SendMessage(LOGWND, EM_SETSEL, sizeBuffer, sizeBuffer);
      		free(buffer);
    		}
  		}
	}

LRESULT CALLBACK OFNHookProc(HWND hdlg,UINT uiMsg,WPARAM wParam,LPARAM lParam)
	{
    switch (uiMsg)
       {
       case WM_INITDIALOG:
         {
         CHAR msg[32];

         wsprintf(msg,"WM_INITDIALOG:%08lx\r\n",GetCapture());
         AppendToLog(msg);
         }
         break;
       case WM_NOTIFY:
         {
         CHAR msg[64];
         LPOFNOTIFY lpn = (LPOFNOTIFY)lParam;
         HWND hCapture;

         hCapture = GetCapture();
         wsprintf(msg,"WM_NOTIFY:%08lx - %08lx - (%i)\r\n",hCapture,
         	lpn->hdr.code,COUNT);
         AppendToLog(msg);
         // ok((COUNT == 2 && hCapture != NULL),"CommonFileDialog - mouse capture should have been released now\n" );
         if (COUNT == 2 && hCapture != NULL)
            {
            AppendToLog("CommonFileDialog - mouse capture should have been released now!\r\n");
            }
         if (COUNT >= 2)
         	{
            AppendToLog("Quitting dialog\r\n");
            PostMessage(GetParent(hdlg),WM_COMMAND,IDCANCEL,IDCANCEL);
            }
         COUNT++;
         }
         break;
       }
	return 0; /* do not process any messages */
	}

LRESULT CALLBACK CFHookProc(HWND hdlg,UINT uiMsg, WPARAM wParam,LPARAM lParam)
	{
    switch (uiMsg)
       {
       case WM_INITDIALOG:
         {
         CHAR msg[32];

         wsprintf(msg,"WM_INITDIALOG:%08lx\r\n",GetCapture());
         AppendToLog(msg);
         }
         break;
       case  WM_CTLCOLORDLG:
         {
         CHAR msg[64];
         HWND hCapture;

         hCapture = GetCapture();
         // ok((COUNT == 2 && hCapture != NULL),"ChooseFontDialog - mouse capture should have been released now\n" );
         wsprintf(msg,"WM_CTLCOLORDLG:%08lx - (%i)\r\n",hCapture,COUNT);
         AppendToLog(msg);
         if (hCapture != NULL)
            {
            AppendToLog("ChooseFontDialog - mouse capture should have been released now!\r\n");
            }
         AppendToLog("Quitting dialog\r\n");
         PostMessage(hdlg,WM_COMMAND,IDCANCEL,IDCANCEL);
         }
         break;
       }
	return 0; /* do not process any messages */
	}

LRESULT CALLBACK MainWndProc(HWND hWnd, UINT message,WPARAM wParam, LPARAM lParam)
	{
	switch (message)
		{
		case WM_CLOSE:
			CloseWindow(hWnd);
			PostQuitMessage(0);
			break;

        case WM_SHOWWINDOW:
            {
            OPENFILENAME opfn;
            CHOOSEFONT chsfnt;
            CHAR fnName[MAX_PATH];
            LOGFONT lFont = {-13,0,0,0,0,0,0,0,0,
           OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
           DEFAULT_QUALITY,FF_ROMAN,"Arial"};
            CHAR msg[32];

            if (wParam == TRUE)
            	{
                AppendToLog("Testing GetOpenFileName\r\n");
            	SetCapture(CAPTWND);
            	wsprintf(msg,"Wnd with capture:%08lx\r\n",GetCapture());
                AppendToLog(msg);

	        	memset(&opfn, 0, sizeof(OPENFILENAME));
	        	memset(fnName,0,sizeof(fnName));

            	opfn.lStructSize = sizeof (OPENFILENAME) ;
            	opfn.hwndOwner = hWnd;
            	opfn.lpstrFile = &fnName;
            	opfn.nMaxFile = MAX_PATH;
                opfn.Flags = OFN_EXPLORER|OFN_ENABLEHOOK ;
                opfn.lpfnHook = OFNHookProc;
            	GetOpenFileName(&opfn);

            	wsprintf(msg,"Wnd with capture:%08lx\r\n",GetCapture());
                AppendToLog(msg);
                ReleaseCapture();

                /* ----*/
                COUNT = 0;
                AppendToLog("Testing ChooseFont\r\n");
            	SetCapture(CAPTWND);
            	wsprintf(msg,"Wnd with capture:%08lx\r\n",GetCapture());
                AppendToLog(msg);

                memset(&chsfnt, 0, sizeof(CHOOSEFONT));
            	chsfnt.lStructSize = sizeof (CHOOSEFONT) ;
            	chsfnt.hwndOwner = hWnd;
            	chsfnt.lpLogFont = &lFont;
            	chsfnt.iPointSize = 10;
                chsfnt.Flags = CF_SCREENFONTS|CF_ENABLEHOOK ;
                chsfnt.lpfnHook = CFHookProc;
                chsfnt.nFontType = REGULAR_FONTTYPE;
                ChooseFont(&chsfnt);

            	wsprintf(msg,"Wnd with capture:%08lx\r\n",GetCapture());
                AppendToLog(msg);
                ReleaseCapture();
            	}
            }
		default:
			return (DefWindowProc(hWnd, message, wParam, lParam));
		}
	return (NULL);
	}

BOOL InitInstance(HANDLE hInstance,int nCmdShow)
	{
	RECT cRect;
	WNDCLASS mwc;

	hInst = hInstance;

	mwc.style = NULL;
	mwc.lpfnWndProc = (WNDPROC)MainWndProc;
	mwc.cbClsExtra = 0;
	mwc.cbWndExtra = 0;
	mwc.hInstance = hInstance;
	mwc.hIcon = NULL;
	mwc.hCursor = NULL;
	mwc.hbrBackground = NULL;
	mwc.lpszMenuName = NULL;
	mwc.lpszClassName = "TEST";
	if (!RegisterClass(&mwc))
		return(FALSE);

	cRect.left = 100;
	cRect.right = cRect.left + 300;
	cRect.top = 100;
	cRect.bottom = cRect.top + 200;
	MAINWND = CreateWindow(
		"TEST",
		"TEST",
		WS_OVERLAPPEDWINDOW,
		cRect.left,
		cRect.top,
		cRect.right,
		cRect.bottom,
		NULL,
		NULL,
		hInstance,
		NULL);

	if (!MAINWND)
		return (FALSE);

    GetClientRect(MAINWND,&cRect);
	LOGWND = CreateWindow(
		"EDIT",
		"",
		WS_CHILD|WS_VISIBLE|ES_MULTILINE,
		cRect.left,
		cRect.top,
		cRect.right,
		cRect.bottom,
		MAINWND,
		NULL,
		hInstance,
		NULL);

	if (!MAINWND)
		return (FALSE);


	CAPTWND = CreateWindow(
		"TEST",
		"TEST",
		WS_OVERLAPPEDWINDOW,
		0,
		0,
		0,
		0,
		NULL,
		NULL,
		hInstance,
		NULL);

	if (!CAPTWND)
		return (FALSE);
	ShowWindow(MAINWND, nCmdShow);
	ShowWindow(CAPTWND, FALSE);
	return (TRUE);
	}


/****************************************************************************

	FUNCTION: WinMain(HANDLE, HANDLE, LPSTR, int)

	PURPOSE: calls initialization function, processes message loop

****************************************************************************/

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpCmdLine,
		int nCmdShow)
	{
	MSG msg;

	if (!InitInstance(hInstance, nCmdShow))
		return (FALSE);

	while (GetMessage(&msg, NULL, NULL, NULL))
		{
        TranslateMessage(&msg);
		DispatchMessage(&msg);
        }
    DestroyWindow(MAINWND);
    DestroyWindow(CAPTWND);
	return (msg.wParam);
	}



